#ifndef VERTEX_PROPERTY_WRITER_H_INCLUDED_
#define VERTEX_PROPERTY_WRITER_H_INCLUDED_

#include "rose.h"

#include "segment_graph.h"
#include "../internal_representation/design.h"
#include "../internal_representation/module_instance.h"
#include "../internal_representation/port.h"
#include "../internal_representation/object.h"
#include "../tools/variable_helper.h"
namespace risc {

namespace sg {

struct VertexPropertyWriter {

  static bool print_read_write_access;

  VertexPropertyWriter(Graph& graph);

  template <class Vertex>
  void operator() (std::ostream &out, Vertex v)
  {
    Segment &seg = graph_[v];
    if(PSG_MODE) {
      write_segment_PSG_mode(out, seg);
    } else if(print_read_write_access) {
      write_segment_RW_mode(out, seg);
    } else {
      write_segment(out, seg);
    }
  }

  /**
   * \fn      write_segment
   * \brief   This function generates DOT code for a segment
   *          (in normal mode)
   */
  void write_segment(std::ostream &out, Segment &seg);

  /**
   * \fn      write_segment
   * \brief   This function generates DOT code for a segment
   *          in read-write variables mode
   */
  void write_segment_RW_mode(std::ostream &out, Segment &seg);

  /**
   * \fn      write_segment_PSG_mode
   * \brief   This function generates DOT code for a segment
   *          in PSG mode
   */
  void write_segment_PSG_mode(std::ostream &out, Segment &seg);

  /**
   * \fn       replace_escape_characters
   * \brief    This function replaces characters with escape sequences.
   */
  static void replace_escape_characters(std::string &unparsed_stmt);

  /**
   * \fn       replace_escape_characters
   * \brief    This function replaces characters with escape sequences.
   */
  static std::string write_pcp(PortCallPath &pcp);

  Graph& graph_;

  friend std::ostream& operator<< (std::ostream &out,
                           const VertexPropertyWriter &vpw)
  {
    return out;
  }
};

}; // end of namespace sg

}; // end of namespace risc

#endif /* VERTEX_PROPERTY_WRITER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
