// bug20180723a.cc:
// (was code_generation_error.cpp reported by DM 07/23/18)

#include <systemc.h>

class write_if : virtual public sc_interface {
    public:
      virtual void write(char) = 0;
};
 
class read_if : virtual public sc_interface {
  public:
    virtual void read(char &) = 0;
};

class fifo : public sc_channel, public write_if, public read_if {
  public: 
    fifo(sc_module_name name) ;
    void write(char c) ;
    void read(char &c) ;

  public:
    enum e { max = 2 };
    char data[max];
    int num_elements, first;
    sc_event write_event, read_event;
};

fifo::fifo(sc_module_name name) : 
  sc_channel(name), 
  num_elements(0), 
  first(0) 
{ }
 
void fifo::write(char c) {
  if (num_elements == max)
    wait(read_event);
  data[(first + num_elements) % max] = c;
  ++ num_elements;
  write_event.notify(SC_ZERO_TIME);
}

void fifo::read(char &c) {
  if (num_elements == 0)
    wait(write_event);
  c = data[first];
  -- num_elements;
  first = (first + 1) % max;
  read_event.notify(SC_ZERO_TIME);
}

class consumer : public sc_module {
  public:
    sc_port<read_if> in;

    SC_HAS_PROCESS(consumer);

    consumer(sc_module_name name) : sc_module(name) {
      SC_THREAD(loop);
    }

    void loop() {
      char c;
        in->read(c);	// RD: sibling calls into the channel via a port (PCP)
        cout << ">> consumer: " << c << endl;
    }
};

class top : public sc_module {
public:
  fifo fifo_inst;
  consumer cons_inst;

  SC_HAS_PROCESS(top);
  top(sc_module_name name) : 
    sc_module(name),
    fifo_inst("Fifo1"),
    cons_inst("Consumer1")
  {
    cons_inst.in.bind(fifo_inst);
    SC_THREAD(behav);
  }

  void behav() {
    char c = 'r';
    fifo_inst.write(c);	// RD: parent calls directly into a child instance!
  }
};

int sc_main (int, char *[]) {
  top top1("Top1");
  sc_start();
  return 0;
}
